"""Graph of a design problem complexity study. Components are related (have edges) to
attributes that are of importance, which are of edge kind "incidence". Nodes have kind
"component" or "attribute".

Reference: [Figure 1(a)] Chen, L., & Li, S. (2005). Analysis of Decomposability and
Complexity for Design Problems in the Context of Decomposition. Journal of Mechanical
Design, 127(4), 545. https://doi.org/10.1115/1.1897405
"""



"""Example 8 node directed graph from Robert Tarjan.

Reference:
[Figure 3] Tarjan, R. (1972). Depth-First Search and Linear Graph Algorithms.
SIAM Journal on Computing, 1(2), 146–160. https://doi.org/10.1137/0201010
"""

edge_weights = ["adjacency"]



"""Graph describing a design problem using objects that need to be designed based on
certain aspects. It is part of a set of three design problem graphs, each ideally solved
with a different approach -- being a **modular**, integral or mixed architecture.

This particular graph describes a design problem which is ideally solved using a
**modular object architecture**. That means that the objects are best put into modules.
Modules on different aspect domains can then be mixed and matched so achieve a solution
for it's combined members. Each module should be designed based on all aspects that the
objects members posess combined, while taking into account the shared dependencies with
other domains. A module's design solution should be applicable to every member.

Nodes are of kind "object" or "aspect_1", "aspect_2", and "aspect_3". The numbers
represent different aspect domains. An "incidence" kind edge from an object to an aspect
means that it posesses that aspect.

Authors: Tim Wilschut and Tiemen Schuijbroek, October 2019.
"""



"""Delta analysis module."""

from typing import Dict, List

from ragraph.analysis.comparison.utils import (
    get_common_leafs,
    get_unique_edges,
    get_unique_leafs,
)
from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node


class DeltaAnalysis:
    """Basic class for performing a delta analysis on two Graph objects.

    Arguments:
        graph_a: Base graph object for comparison.
        graph_b: Reference graph object for comparison.

    Note:
        Graphs are compared on leaf node level, where node names are used to determine
        if nodes are unique or not.
    """

    def __init__(self, graph_a: Graph, graph_b: Graph):
        self.graph_a = graph_a
        self.graph_b = graph_b

    def _reset_lazy_props(self):
        """Reset all lazy properties."""
        self._unq_leafs_a = None
        self._unq_leafs_b = None
        self._sm_leafs_ab = None
        self._cm_leafs_ab = None
        self._unq_edges_a = None
        self._unq_edges_b = None
        self._cm_edges_ab = None
        self._node_dict = None
        self._nodes = []
        self._node_dict = None
        self._edges = []

    @property
    def graph_a(self) -> Graph:
        """Graph A to be compared."""
        return self._graph_a

    @graph_a.setter
    def graph_a(self, value):
        if not isinstance(value, Graph):
            raise TypeError(f"Key '{value}' should be a Graph object.")
        self._graph_a = value
        self._reset_lazy_props()

    @property
    def graph_b(self) -> Graph:
        """Graph B to be compared."""
        return self._graph_b

    @graph_b.setter
    def graph_b(self, value):
        if not isinstance(value, Graph):
            raise TypeError(f"Key '{value}' should be a Graph object.")
        self._graph_b = value
        self._reset_lazy_props()

    @property
    def unq_leafs_a(self) -> List[Node]:
        """List of leafs that are only present in graph a."""
        if self._unq_leafs_a is None:
            self._unq_leafs_a = get_unique_leafs(a=self.graph_a, b=self.graph_b)
        return self._unq_leafs_a

    @property
    def unq_leafs_b(self) -> List[Node]:
        """List of leafs that are only present in graph b."""
        if self._unq_leafs_b is None:
            self._unq_leafs_b = get_unique_leafs(a=self.graph_b, b=self.graph_a)
        return self._unq_leafs_b

    @property
    def cm_leafs_ab(self) -> List[Node]:
        """List of leafs that are present in graph a and b."""
        if self._cm_leafs_ab is None:
            ue = self.unq_edges_a + self.unq_edges_b
            self._cm_leafs_ab = get_common_leafs(a=self.graph_a, b=self.graph_b, ue=ue)
        return self._cm_leafs_ab

    @property
    def unq_edges_a(self) -> List[Edge]:
        """List of edges that are only present in graph a."""
        if self._unq_edges_a is None:
            self._unq_edges_a = get_unique_edges(a=self.graph_a, b=self.graph_b)
        return self._unq_edges_a

    @property
    def unq_edges_b(self) -> List[Edge]:
        """List of edges that are only present in graph b."""
        if self._unq_edges_b is None:
            self._unq_edges_b = get_unique_edges(a=self.graph_b, b=self.graph_a)
        return self._unq_edges_b

    @property
    def cm_edges_ab(self) -> List[Edge]:
        """List of edges that are present in both graphs."""
        if self._cm_edges_ab is None:
            self._cm_edges_ab = [
                edge for edge in self.graph_a.edge_list if edge not in set(self.unq_edges_a)
            ]
        return self._cm_edges_ab

    @property
    def node_dict(self) -> Dict[str, Node]:
        """Dictionary from node name to Node object"""
        if self._node_dict is None:
            self._node_dict = {n.name: n for n in self.nodes}

        return self._node_dict

    @property
    def nodes(self) -> List[Node]:
        """List of nodes of the delta graph."""
        if not self._nodes:
            for leaf in self.unq_leafs_a:
                self._add_node(n=leaf, kind=leaf.kind + "_unique_a")
            for leaf in self.cm_leafs_ab:
                self._add_node(n=leaf, kind=leaf.kind + "_common_ab")
            for leaf in self.unq_leafs_b:
                self._add_node(n=leaf, kind=leaf.kind + "_unique_b")

        return self._nodes

    @property
    def edges(self) -> List[Edge]:
        """List of edges of the delta graph."""
        if not self._edges:
            for edge in self.unq_edges_a:
                self._add_edge(e=edge, kind=edge.kind + "_unique_a")
            for edge in self.cm_edges_ab:
                self._add_edge(e=edge, kind=edge.kind + "_common_ab")
            for edge in self.unq_edges_b:
                self._add_edge(e=edge, kind=edge.kind + "_unique_b")

        return self._edges

    def _add_node(self, n: Node, kind: str) -> None:
        """Add Node to Delta Graph."""
        self._nodes.append(
            Node(
                name=n.name,
                kind=kind,
                labels=n.labels,
                weights=n.weights,
                annotations=n.annotations,
            )
        )

    def _add_edge(self, e: Edge, kind: str) -> None:
        """Add edge to Delta Graph."""
        self._edges.append(
            Edge(
                source=self.node_dict[e.source.name],
                target=self.node_dict[e.target.name],
                kind=kind,
                labels=e.labels,
                weights=e.weights,
                annotations=e.annotations,
            )
        )



"""Sigma analysis module. """

from collections import defaultdict
from typing import Dict, List, Tuple, Union

from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node


class SigmaAnalysis:
    """Basic class to perform a sigma analysis on :obj:`ragraph.graph.Graph` objects.

    Arguments:
        graphs: List of graphs to compare. mode: One of "absolute" or "fraction".
        mode: Determines whether edge label overlap should be calculated as an absolute
            count over all edges between certain nodes or whether a fraction (percentage
            of label occurrence) is calculated as edge weights. Defaults to fraction.

    Note:
        Graphs are compared on leaf node level. The resulting sigma or overlap is
        currently only calculated in terms of overlapping edge labels.

    Note:
        Apart from label occurrence, the overall edge count is also added as an Edge
        weight, named `"_count"`. (The underscore prefix is added so it doesn't
        collide with regular label names)
    """

    def __init__(self, graphs: List[Graph], mode: str = "fraction"):
        self.graphs = graphs
        self.mode = mode

    def _reset_lazy_props(self):
        """Reset all lazy properties."""
        self._nodes = []
        self._node_dict = None
        self._edges = []
        self._edge_dict = None

    @property
    def graphs(self) -> List[Graph]:
        """List of graphs to be compared."""
        return self._graphs

    @graphs.setter
    def graphs(self, value):
        if not isinstance(value, list):
            raise TypeError(f"Key '{value}' should be a list of Graph objects.")

        for v in value:
            if not isinstance(v, Graph):
                raise TypeError(f"Key '{v}' should be a Graph object.")

        self._graphs = value
        self._reset_lazy_props()

    @property
    def mode(self) -> str:
        """Edge weight mode."""
        return self._mode

    @mode.setter
    def mode(self, value):
        if value not in ["fraction", "absolute"]:
            raise ValueError(f"'{value}' should be one of 'fraction' or 'absolute'.")

        self._mode = value
        self._reset_lazy_props()

    @property
    def nodes(self) -> List[Node]:
        """List of leaf nodes within the graph."""
        if not self._nodes:
            node_dict = {n.name: n for g in self.graphs for n in g.leafs}
            for n in node_dict.values():
                self._nodes.append(
                    Node(
                        name=n.name,
                        kind=n.kind,
                        labels=n.labels,
                        weights=n.weights,
                        annotations=n.annotations,
                    )
                )

        return self._nodes

    @property
    def node_dict(self) -> Dict[str, Node]:
        """Dictionary of node name to node."""
        if not self._node_dict:
            self._node_dict = {n.name: n for n in self.nodes}

        return self._node_dict

    @property
    def edges(self) -> List[Edge]:
        if not self._edges:
            edge_dict = defaultdict(list)

            for g in self._graphs:
                for e in g.edges:
                    edge_dict[e.source.name, e.target.name, e.kind].append(e)

            for key, edges in edge_dict.items():
                self._edges.append(
                    Edge(
                        source=self.node_dict[key[0]],
                        target=self.node_dict[key[1]],
                        kind=key[2],
                        labels=sorted(set([label for e in edges for label in e.labels])),
                        weights=self._get_edge_weights(edges),
                    )
                )

        return self._edges

    @property
    def edge_dict(self) -> Dict[Tuple[str, str], List[Edge]]:
        """Dictionary of source and target names to list of edges."""
        if not self._edge_dict:
            self._edge_dict = defaultdict(list)
            for e in self.edges:
                self._edge_dict[e.source.name, e.target.name].append(e)

        return self._edge_dict

    def _get_edge_weights(self, edges: List[Edge]) -> Dict[str, Union[int, float]]:
        """Compute weights for edge in Sigma DSM from the provided list of edges."""
        weight_counts: Dict[str, Union[int, float]] = {"_count": len(edges)}

        for e in edges:
            for label in e.labels:
                weight_counts[label] = weight_counts.get(label, 0) + 1

        div = 1 / len(self.graphs)
        return (
            weight_counts
            if self.mode == "absolute"
            else {key: value * div for key, value in weight_counts.items()}
        )



"""
###################
Similarity analyses
###################

Graph similarity is often expressed as a metric, where nodes and edges are scanned for
similar patterns, properties, or other aspects. There are three levels of equivalence,
being structural, automorphic, or regular equivalence. Where each of the former implies
all latter equivalences, respectively.


******************
Available analyses
******************

The following algorithms are directly accessible after importing
:obj:`ragraph.analysis.similarity`:

* :obj:`jaccard_index`: Jaccard Similarity Index of two objects based on the number
  properties they both possess divided by the number of properties either of them have.
* :obj:`jaccard_matrix`: Jaccard Similarity Index between a set of objects stored in a
  square matrix.

.. note:: Both Jaccard methods require a callable that takes an object and returns a
   list of booleans representing the possession of a property (the :obj:`on` argument).
   Some examples are included in the :mod:`ragraph.analysis.similarity.utils` module,
   like :func:`ragraph.analysis.similarity.utils.on_hasattrs`.
"""
# flake8: noqa, ignore errors on unused imports here.

from ragraph.analysis.similarity._jaccard import jaccard_index, jaccard_matrix
from ragraph.analysis.similarity._similarity import SimilarityAnalysis
